#!/usr/bin/env python3
# -*- coding:utf-8 -*-

import sys
import json
import click

from typing import Any, Dict, Optional, Tuple, Type, Union
from pathlib import Path
from colort import colorize, ForegroundColor, BackgroundColor
from up2b.up2b_lib.up2b_api import choose_image_bed
from up2b.up2b_lib.up2b_api.sm import SM
from up2b.up2b_lib.up2b_api.imgtu import Imgtu
from up2b.up2b_lib.up2b_api.github import Github
from up2b.up2b_lib.up2b_api.imgtg import Imgtg
from up2b.up2b_lib.constants import (
    CACHE_PATH,
    CONF_FILE,
    IMAGE_BEDS_HELP_MESSAGE,
    IMAGE_BEDS_NAME,
    ImageBedCode,
)
from up2b.up2b_lib.utils import check_paths, read_conf
from up2b.up2b_lib.log import logger
from up2b.version import __version__  # Automatically create version.py after building

IMAGE_BEDS: Dict[
    ImageBedCode, Union[Type[SM], Type[Imgtu], Type[Github], Type[Imgtg]]
] = {
    ImageBedCode.SM_MS: SM,
    ImageBedCode.IMGTU: Imgtu,
    ImageBedCode.GITHUB: Github,
    ImageBedCode.IMGTG: Imgtg,
}


def echo(*args: Any):
    click.echo(" ".join([str(i) for i in args]))


CONTEXT_SETTINGS = dict(help_option_names=["-h", "--help"])


@click.group(name="cli", context_settings=CONTEXT_SETTINGS)
@click.help_option("-h", "--help", help="显示本帮助信息")
@click.version_option(
    __version__, "-v", "--version", message="%(version)s", help="显示版本信息"
)
def cli():
    if not CACHE_PATH.exists():
        logger.debug("缓存目录不存在")
        CACHE_PATH.mkdir()
        logger.info("已创建缓存目录", cache_path=CACHE_PATH)


@cli.command(help="显示正在使用的图床")
def current():
    click.echo(
        f"""{colorize(" " + chr(10004), ForegroundColor.GREEN)} {_read_image_bed(False, False)}""",
    )


@cli.command(help="列出支持的所有图床")
def list():
    print_list()


@cli.command(
    short_help="选择图床",
    help=f"""图床代码对应如下：\n\n{IMAGE_BEDS_HELP_MESSAGE}""",
)
@click.help_option("-h", "--help", help="显示本帮助信息")
@click.argument(
    "code",
    nargs=1,
    type=int,
    metavar="<int>",
)
def choose(code: int):
    choose_image_bed(code)
    echo(
        colorize(" ==>", ForegroundColor.CYAN),
        IMAGE_BEDS[ImageBedCode(code)](),
    )


@cli.command(help="保存图床登录信息")
@click.help_option("-h", "--help", help="显示本帮助信息")
@click.argument(
    "username",
    nargs=1,
    type=str,
)
@click.argument(
    "password",
    nargs=1,
    type=str,
)
def login(username: str, password: str):
    ib = _read_image_bed()
    if isinstance(ib, Github):
        logger.fatal(
            "you have chosen `github` as the image bed, please login with `-lg`"
        )

    logger.info("current image bed", name=str(ib))

    echo("正在验证账号，请耐心等待...")

    ok = ib.login(username, password)
    if not ok:
        logger.fatal("username or password incorrect")

    echo("账号验证通过")


@cli.command(help="保存 git 登录信息")
@click.help_option("-h", "--help", help="显示本帮助信息")
@click.argument(
    "access_token",
    nargs=1,
    type=str,
)
@click.argument(
    "username",
    nargs=1,
    type=str,
)
@click.argument(
    "repository",
    nargs=1,
    type=str,
)
@click.argument(
    "path",
    nargs=1,
    type=str,
)
def login_git(access_token: str, username: str, repository: str, path: str):
    ib = _read_image_bed()
    if not isinstance(ib, Github):
        logger.fatal(
            "the image bed you choose is not `github`, , please login with `-l`"
        )

    ib.login(access_token, username, repository, path)


@cli.command(help="上传图片")
@click.argument(
    "image_paths",
    nargs=-1,
    type=click.Path(exists=True, file_okay=True, dir_okay=False),
    metavar="[IMAGE_PATH]...",
)
@click.option(
    "-aw",
    "--add-watermark",
    is_flag=True,
    show_default=True,
    default=False,
    help="对要上传的图片添加文字水印",
)
@click.option(
    "-ac",
    "--auto-compress",
    is_flag=True,
    show_default=True,
    default=False,
    help="允许自动压缩图片",
)
@click.option(
    "-ic",
    "--ignore-cache",
    is_flag=True,
    show_default=True,
    default=False,
    help="忽略数据库缓存，强制上传图片",
)
@click.option(
    "-q",
    "--quiet",
    is_flag=True,
    show_default=True,
    default=False,
    help="静默模式。开启后不显示上传进度",
)
@click.option("-t", "--timeout", type=float, help="上传图片的超时时间")
def upload(
    image_paths: Tuple[str],
    add_watermark: bool,
    auto_compress: bool,
    ignore_cache: bool,
    timeout: float,
    quiet: bool,
):
    ib = _read_image_bed(
        add_watermark=add_watermark,
        auto_compress=auto_compress,
        ignore_cache=ignore_cache,
        timeout=timeout,
        quiet=quiet,
    )

    paths = check_paths(image_paths)

    ib.upload_images(*paths)


@cli.command(
    short_help="手动添加缓存", help="某些图床禁止上传重复图片，但在上传时又不返回旧图片地址，此时需要你手动获取原图片地址并执行此命令添加到缓存中。"
)
@click.argument(
    "image_path",
    type=click.Path(exists=True, file_okay=True, dir_okay=False, path_type=Path),
)
@click.argument("url", type=str)
def add_cache(image_path: Path, url: str):
    ib = _read_image_bed()
    ib.cache.add(image_path, url, IMAGE_BEDS_NAME[ib.image_bed_code])


@cli.command(
    short_help="配置文字水印",
    help="""
    参数解释：

        - x 整数，横坐标
    
        - y 整数，纵坐标
    
        - opacity 整数，透明度。50 即为透明度 50%。
    
        - text 字符串，水印文字内容
    
        - font_path 字符串，字体路径
    
        - size 整数，字体大小
    """,
)
@click.argument(
    "x",
    nargs=1,
    type=int,
)
@click.argument(
    "y",
    nargs=1,
    type=int,
)
@click.argument(
    "opacity",
    nargs=1,
    type=int,
)
@click.argument(
    "text",
    nargs=1,
    type=str,
)
@click.argument(
    "font_path",
    nargs=1,
    type=str,
)
@click.argument(
    "size",
    nargs=1,
    type=int,
)
def config_watermark(
    x: int, y: int, opacity: int, text: str, font_path: str, size: int
):
    _config_text_watermark(x, y, opacity, text, font_path, size)


def _read_image_bed(
    auto_compress: bool = False,
    add_watermark: bool = False,
    ignore_cache: bool = False,
    timeout: Optional[float] = None,
    quiet: bool = False,
) -> Union[SM, Imgtu, Imgtg, Github]:
    conf = read_conf()

    selected_code = conf.image_bed
    if selected_code == None:
        logger.fatal("当前图床为空，请先选择要使用的图床")

    assert isinstance(selected_code, int)

    try:
        code = ImageBedCode(selected_code)

        return IMAGE_BEDS[code](
            auto_compress=auto_compress,
            add_watermark=add_watermark,
            ignore_cache=ignore_cache,
            timeout=timeout,
            quiet=quiet,
        )
    except ValueError:
        IMAGE_BEDS[ImageBedCode.SM_MS](
            auto_compress=auto_compress,
            add_watermark=add_watermark,
            ignore_cache=ignore_cache,
            timeout=timeout,
            quiet=quiet,
        )
        logger.fatal("未知的图床代码，可能因为清除无效的 gitee 配置，请重试", code=selected_code)


def _config_text_watermark(
    x: int, y: int, opacity: int, text: str, font: str, size: int
):
    try:
        with open(CONF_FILE, "r+") as f:
            conf = json.loads(f.read())
            f.seek(0, 0)
            conf["watermark"] = {
                "x": x,
                "y": y,
                "opacity": opacity,
                "text": text,
                "font": font,
                "size": size,
            }
            f.write(json.dumps(conf))
            f.truncate()
    except FileNotFoundError:
        with open(CONF_FILE, "w") as f:
            f.write(
                json.dumps(
                    {
                        "watermark": {
                            "x": x,
                            "y": y,
                            "opacity": opacity,
                            "text": text,
                            "font": font,
                            "size": size,
                        }
                    }
                )
            )


def print_list() -> int:
    conf = read_conf()
    auth_data = conf.auth_data

    assert isinstance(auth_data, dict)

    if not auth_data:
        if conf.image_bed is None:
            logger.warning(
                "no image bed selected, "
                + "you need to use `--choose-site` or `-c` to select the image bed first."
            )

        if conf.image_bed:
            logger.warning(
                "no authentication information has been configured",
                image_bed=colorize(
                    IMAGE_BEDS[conf.image_bed](conf=conf).__str__(),
                    BackgroundColor.DARK_GRAY,
                    ForegroundColor.WHITE,
                ),
                code=conf.image_bed,
            )

    def print_item(symbol, color, idx):
        ib = IMAGE_BEDS[ImageBedCode(idx)](conf=conf)
        click.echo(f"""{colorize(" " + symbol, color)} {idx} {ib}: {ib.description}""")

    for i in ImageBedCode:
        if auth_data.get(i):
            print_item(chr(10003), ForegroundColor.GREEN, i)
        else:
            print_item(chr(10007), ForegroundColor.RED, i)

    return 0


def run_main():
    sys.exit(cli())


if __name__ == "__main__":
    run_main()



#!/usr/bin/env python3
# -*- coding:utf-8 -*-
# @Author:      thepoy
# @Email:       thepoy@163.com
# @File Name:   server.py
# @Created At:  2023-02-08 14:18:22
# @Modified At: 2023-02-21 14:02:33
# @Modified By: thepoy

import socketserver
import json
import re

from http.server import BaseHTTPRequestHandler
from pathlib import Path
from typing import Any, Dict, Optional
from tests import IMAGES_DIR
from requests_toolbelt.multipart import decoder


filename_ptn = re.compile(rb'filename="(.+?)"')


HOST = "127.0.0.1"
PORT = 8080
ADDR = (HOST, PORT)


class Handler(BaseHTTPRequestHandler):
    def set_headers(
        self,
        content_type: str,
        content_length: int,
        others: Optional[Dict[str, Any]] = None,
    ):
        self.send_header("Content-type", content_type)
        self.send_header("Content-Length", str(content_length))

        if others:
            for k, v in others.items():
                self.send_header(k, v if isinstance(v, str) else str(v))

        self.end_headers()

    def do_GET(self):
        path = Path("." + self.path)

        print(path)

        if not path.exists():
            self.send_error(404)

            return

        self.send_response(200)

        self.set_headers("image/png", path.stat().st_size)

        self.wfile.write(path.open("rb").read())

    def do_POST(self):
        length = int(self.headers["content-length"])
        body = self.rfile.read(length)

        content_type = self.headers["Content-Type"]

        part = decoder.MultipartDecoder(body, content_type).parts[0]

        disposition: bytes = part.headers[b"Content-Disposition"]  # type: ignore
        searched = filename_ptn.search(disposition)
        if not searched:
            self.send_error(400)

            return

        filename = searched.group(1).decode()
        filepath = IMAGES_DIR / filename

        if not filepath.exists():
            self.send_error(400)

            return

        if filepath.stat().st_size != len(part.content):
            self.send_error(400, "upload failed")

            return

        resp = {
            "status": "ok",
            "url": f"http://{HOST}:{PORT}/images/{filename}",
        }

        self.send_response(200)
        self.send_header("Content-type", "application/json")
        self.end_headers()
        self.wfile.write(json.dumps(resp).encode())
        return


def check_server():
    import socket

    sk = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
    sk.settimeout(1)

    while True:
        try:
            sk.connect(ADDR)
            break
        except Exception:
            continue

    sk.close()


def run_server():
    with socketserver.TCPServer(ADDR, Handler) as httpd:
        print("serving at port", PORT)
        httpd.serve_forever()


if __name__ == "__main__":
    run_server()



#!/usr/bin/env python3
# -*- coding:utf-8 -*-
# @Author:      thepoy
# @Email:       thepoy@163.com
# @File Name:   __init__.py
# @Created At:  2021-02-08 16:00:23
# @Modified At: 2023-02-21 12:41:46
# @Modified By: thepoy



#!/usr/bin/env python3
# -*- coding:utf-8 -*-

import json
import os
import locale
import requests

from typing import List, Sequence, Tuple, Union
from functools import wraps, partial
from pathlib import Path
from concurrent.futures import ThreadPoolExecutor, as_completed
from urllib.parse import urlparse

from up2b.up2b_lib.constants import (
    CONF_FILE,
    DEFAULT_TIMEOUT,
    IS_MACOS,
    PYTHON_VERSION,
    CACHE_PATH,
    ImageBedCode,
)
from up2b.up2b_lib.custom_types import (
    Config,
    DownloadErrorResponse,
    ImageType,
)
from up2b.up2b_lib.log import child_logger


logger = child_logger(__name__)


def timeout_in_env() -> float:
    t = os.getenv("UP2B_TIMEOUT")
    if not t:
        return DEFAULT_TIMEOUT

    try:
        return float(t)
    except ValueError:
        return DEFAULT_TIMEOUT


def check_image_exists(images: Tuple[Union[ImageType, DownloadErrorResponse], ...]):
    for image in images:
        if isinstance(image, Path) and not image.exists():
            raise FileNotFoundError(image)


def read_conf() -> Config:
    logger.trace("reading configuration file")

    if os.getenv("UP2B_TEST"):
        return Config(None, {}, {})

    if not CONF_FILE.exists():
        logger.warning(
            "the configuration file is not found, "
            + "you need to use `choose` command to select the image bed first."
        )

        return Config(None, {}, {})

    try:
        with CONF_FILE.open(encoding="utf-8") as f:
            conf = json.loads(f.read())
    except Exception as e:
        logger.fatal("unkown error", err=e)
    else:
        auth_data = {
            ImageBedCode(int(k)): v for k, v in conf.get("auth_data", {}).items()
        }

        return Config(
            ImageBedCode(conf.get("image_bed")), auth_data, conf.get("watermark", {})
        )


class Login:
    def __init__(self, func):
        wraps(func)(self)

    def __call__(self, instance, *args, **kwargs):
        return self.__wrapped__(instance, *args, **kwargs)  # type: ignore

    def __get__(self, instance, cls):
        if not instance:
            return partial(self, instance)

        if not instance.auth_info:
            logger.fatal(
                "you have not logged in yet, please use the `-l` or `--login` parameter to log in"
                + " first.\nCurrent image bed code is : `%s`.",
                code=instance.image_bed_code,
            )

        return partial(self, instance)


def is_ascii(char: str):
    assert len(char) == 1, "只能处理字符，而不是字符串"
    return ord(char) < 128


def get_default_language() -> str:
    """默认使用中文，反正也没老外用。

    :returns: 语言标识
    :rtype: str
    """
    if IS_MACOS and PYTHON_VERSION < (3, 7, 5):
        logger.warning(
            f"the version [ {PYTHON_VERSION} ] below 3.7.5 on the macOS platform cannot obtain the system locale, and [ en_US ] is used by default."
        )
        lang = None
    else:
        sys_locale = locale.getdefaultlocale()

        logger.debug(f"system locale: {sys_locale}")

        lang = sys_locale[0]

    # return lang if lang else "en_US"
    return lang if lang else "zh_CN"


def is_url(path: str) -> bool:
    if not path.startswith("http"):
        return False

    try:
        urlparse(path)
        return True
    except ValueError:
        return False


def download_online_image(url: str):
    logger.debug("下载在线图片", url=url)

    resp = requests.get(url)
    if resp.status_code != 200:
        logger.error("在线图片下载失败", status_code=resp.status_code, body=resp.text)
        return DownloadErrorResponse(resp.status_code, resp.text)

    filename = os.path.basename(url)

    if not CACHE_PATH.exists():
        CACHE_PATH.mkdir()

    cache_path = CACHE_PATH / filename

    with cache_path.open("wb") as fb:
        fb.write(resp.content)

    logger.debug("在线图片已保存到缓存目录", cache_path=cache_path)

    return cache_path


def check_path(path: Union[Path, str]):
    if not is_url(str(path)):
        logger.debug("本地图片", path=path)
        return Path(path)

    logger.info("在线图片", path=path)

    return download_online_image(str(path))


def check_paths(paths: Union[Sequence[str], Sequence[Path]]):
    if len(paths) == 1:
        return [check_path(Path(paths[0]))]

    new_paths: List[Union[Path, DownloadErrorResponse]] = [Path()] * len(paths)
    logger.info("使用线程池下载多张图片...")
    with ThreadPoolExecutor(4) as pool:
        futures = {
            pool.submit(check_path, Path(paths[i])): i for i in range(len(paths))
        }

        for future in as_completed(futures):
            idx = futures[future]
            data = future.result()
            new_paths[idx] = data

    logger.info("检查结果", original_paths=paths, new_paths=new_paths)

    return new_paths



#!/usr/bin/env python3
# -*- coding:utf-8 -*-
# @Author:      thepoy
# @Email:       thepoy@163.com
# @File Name:   constants.py
# @Created At:  2021-02-13 09:17:07
# @Modified At: 2023-04-19 14:42:31
# @Modified By: thepoy

import sys
import os

from enum import IntEnum
from pathlib import Path
from tempfile import gettempdir

PYTHON_VERSION = sys.version_info


class ImageBedCode(IntEnum):
    SM_MS = 0
    IMGTU = 1
    GITHUB = 2
    # CODING = 3 # 废弃
    IMGTG = 4

    def __repr__(self) -> str:
        return str(self.value)

    def __str__(self) -> str:
        return str(self.value)


DEFAULT_TIMEOUT = 10.0

# fmt: off
IMAGE_BEDS_CODE = {
    "sm.ms":      ImageBedCode.SM_MS,
    "imgse.com":  ImageBedCode.IMGTU,
    "github.com": ImageBedCode.GITHUB,
    "imgtg.com": ImageBedCode.IMGTG,
}
# fmt: on


IMAGE_BEDS_NAME = {v: k for k, v in IMAGE_BEDS_CODE.items()}

IMAGE_BEDS_HELP_MESSAGE = "\n\n".join([f"{v}: {k}" for k, v in IMAGE_BEDS_CODE.items()])


IS_WINDOWS = sys.platform == "win32"
IS_MACOS = sys.platform == "darwin"

CONFIG_FOLDER_PATH = (
    Path(os.environ["APPDATA"]) / "up2b"
    if IS_WINDOWS
    else Path(os.environ["HOME"]) / ".config" / "up2b"
)

if not CONFIG_FOLDER_PATH.exists():
    os.makedirs(CONFIG_FOLDER_PATH / "conf", 0o755)

CONF_FILE = CONFIG_FOLDER_PATH / "conf" / "conf.up2b.json"
CACHE_PATH = Path(gettempdir()) / "up2b"
CACHE_DATABASE = CONFIG_FOLDER_PATH / "cache.db"



#!/usr/bin/env python3
# -*- coding:utf-8 -*-
# @Author:      thepoy
# @Email:       thepoy@163.com
# @File Name:   errors.py
# @Created At:  2021-02-08 17:04:25
# @Modified At: 2023-02-21 12:41:08
# @Modified By: thepoy


class UploadFailed(Exception):
    pass


class OverSizeError(ValueError):
    pass


class NotLogin(NotImplementedError):
    pass


class UnsupportedType(TypeError):
    pass


class MissingAuth(ValueError):
    pass



#!/usr/bin/env python3
# -*- coding:utf-8 -*-

import sys
import requests
import requests_toolbelt

from typing import Any, Dict, Optional
from tqdm import tqdm
from up2b.up2b_lib.file import File


class ProgressBar(tqdm):
    def update_to(self, n: int) -> None:
        self.update(n - self.n)


def upload_with_progress_bar(
    url: str,
    file: File,
    timeout: float,
    form: Optional[Dict[str, Any]] = None,
    headers: Optional[Dict[str, str]] = None,
):
    data = form or {}
    data.update(file.to_dict())

    encoder = requests_toolbelt.MultipartEncoder(data)

    headers = headers or {}

    with ProgressBar(
        total=encoder.len,
        desc=file.filename,
        unit="B",
        unit_scale=True,
        unit_divisor=1024,
        miniters=1,
        file=sys.stdout,
    ) as bar:
        monitor = requests_toolbelt.MultipartEncoderMonitor(
            encoder, lambda monitor: bar.update_to(monitor.bytes_read)
        )

        headers.update({"Content-Type": monitor.content_type})

        resp = requests.post(url, data=monitor, headers=headers, timeout=timeout)

        return resp



#!/usr/bin/env python3

from enum import IntEnum
from typing import Any, Dict, List, Optional, Union
from dataclasses import dataclass, asdict
from pathlib import Path

from up2b.up2b_lib.constants import ImageBedCode


class ImageBedType(IntEnum):
    common = 1
    git = 2


@dataclass
class ImageStream:
    filename: str
    stream: bytes
    mime_type: str

    def __repr__(self) -> str:
        return self.filename


ImagePath = Path
AuthInfo = Dict[str, Any]
AuthData = Dict[str, AuthInfo]
WaterMarkConfig = Dict[str, int]
ConfigFile = Dict[str, Union[int, AuthData, WaterMarkConfig]]
DeletedResponse = Dict[str, Union[bool, str]]
ImageType = Union[ImagePath, ImageStream]
Images = List[ImageType]


@dataclass
class Config:
    image_bed: Optional[ImageBedCode]
    auth_data: Dict[ImageBedCode, Dict[str, str]]
    watermark: WaterMarkConfig


@dataclass
class BaseResponse:
    def to_dict(self) -> Dict[str, Any]:
        return asdict(self)


@dataclass
class ErrorResponse(BaseResponse):
    status_code: int
    error: Union[str, Dict[str, Any]]


@dataclass
class DownloadErrorResponse(ErrorResponse):
    pass


@dataclass
class UploadErrorResponse(ErrorResponse):
    image_path: str


@dataclass
class GitGetAllImagesResponse(BaseResponse):
    url: str
    sha: str
    delete_url: str


@dataclass
class SMMSResponse(BaseResponse):
    url: str
    delete_url: str
    width: int
    height: int


@dataclass
class CodingResponse(BaseResponse):
    url: str
    filename: str


@dataclass
class ImgtuResponse(BaseResponse):
    id: str
    url: str
    display_url: str
    width: int
    height: int



#!/usr/bin/env python3
# -*- coding:utf-8 -*-

from colorful_logger import get_logger, child_logger as cl
from colorful_logger.logger import get_level_from_env
from up2b.up2b_lib.constants import CONFIG_FOLDER_PATH

log_file_path = CONFIG_FOLDER_PATH / "up2b.log"


def child_logger(name: str):
    return cl(name, logger)


logger = get_logger(
    "up2b",
    level=get_level_from_env(),
    file_path=log_file_path,
    add_file_path=False,
    disable_line_number_filter=True,
    asynchronous=False,
)



import mimetypes

from typing import Optional, Tuple, BinaryIO, Union
from pathlib import Path
from up2b.up2b_lib.custom_types import ImageType


class File:
    def __init__(
        self, key: str, image: ImageType, filename: Optional[str] = None
    ) -> None:
        self.__key = key

        self.mime_type = (
            mimetypes.guess_type(image)[0]
            if isinstance(image, Path)
            else image.mime_type
        )

        self.filename = (
            filename or image.name if isinstance(image, Path) else image.filename
        )
        self.image = image

    @property
    def key(self):
        return self.__key

    def to_tuple(self) -> Tuple[str, Union[BinaryIO, bytes], str]:
        return (
            self.filename,
            self.image.open("rb")
            if isinstance(self.image, Path)
            else self.image.stream,
            self.mime_type,  # type: ignore
        )

    def to_dict(self):
        return {self.key: self.to_tuple()}



#!/usr/bin/env python3
# -*- coding:utf-8 -*-

import os
import requests

from base64 import b64encode
from typing import Optional
from up2b.up2b_lib.custom_types import (
    Config,
    ImageBedType,
    ImagePath,
    ImageStream,
    ErrorResponse,
)
from up2b.up2b_lib.up2b_api import GitBase
from up2b.up2b_lib.constants import ImageBedCode
from up2b.up2b_lib.log import child_logger

logger = child_logger(__name__)


class Github(GitBase):
    image_bed_code = ImageBedCode.GITHUB
    max_size = 20 * 1024 * 1024
    api_url = "https://api.github.com"
    image_bed_type = ImageBedType.git

    def __init__(
        self,
        auto_compress: bool = False,
        add_watermark: bool = False,
        ignore_cache: bool = False,
        conf: Optional[Config] = None,
        timeout: Optional[float] = None,
        quiet: bool = False,
    ):
        super().__init__(
            auto_compress, add_watermark, ignore_cache, conf, timeout, quiet
        )

        if hasattr(self, "token"):
            self.headers = {
                "Accept": "application/vnd.github.v3+json",
                "Authorization": "token " + self.token,
            }

    def upload_image(self, image_path: ImagePath):
        basename = os.path.basename(image_path)
        logger.debug("uploading", name=basename)
        with open(image_path, "rb") as fb:
            data = {
                "content": b64encode(fb.read()).decode("utf-8"),
                "message": "up2b - " + basename,
            }
        return self._upload(image_path, data)

    def upload_image_stream(self, image: ImageStream):
        logger.debug("uploading", name=image.filename)
        data = {
            "content": b64encode(image.stream).decode("utf-8"),
            "message": "up2b - " + image.filename,
        }
        return self._upload(image, data)

    def _get_all_images_in_image_bed(self):
        resp = requests.get(self.base_url, headers=self.headers, timeout=self.timeout)

        return resp

    def delete_image(
        self,
        sha: str,
        url: str,
        message: str = "Delete pictures that are no longer used",
    ) -> Optional[ErrorResponse]:
        return self._delete_image(sha, url, message)

    def cdn_url(self, url: str) -> str:
        path = url.split("/main/")[-1]
        return "https://cdn.jsdelivr.net/gh/%s/%s/%s" % (self.username, self.repo, path)

    def __repr__(self):
        return "github.com"

    @property
    def description(self):
        return "不适合中国大陆使用。访问图片可以用 cdn，但上传图片比较困难。"



#!/usr/bin/env python3
# -*- coding:utf-8 -*-

import os
import re
import time
import json
import requests

from urllib import parse
from pathlib import Path
from typing import Dict, List, Optional, Set, Tuple, Union
from up2b.up2b_lib.custom_types import (
    Config,
    ErrorResponse,
    ImageBedType,
    ImageType,
    ImageStream,
    ImagePath,
    ImgtuResponse,
    UploadErrorResponse,
)
from up2b.up2b_lib.errors import MissingAuth
from up2b.up2b_lib.file import File
from up2b.up2b_lib.http import upload_with_progress_bar
from up2b.up2b_lib.up2b_api import Base
from up2b.up2b_lib.constants import IMAGE_BEDS_NAME, ImageBedCode
from up2b.up2b_lib.log import child_logger

logger = child_logger(__name__)


class Imgtu(Base):
    image_bed_type = ImageBedType.common
    image_bed_code = ImageBedCode.IMGTU
    max_size = 10 * 1024 * 1024
    base_url = "https://imgse.com/"
    __headers = {
        "Accept": "application/json",
        "User-Agent": "Mozilla/5.0 (X11; Linux x86_64; rv:85.0) Gecko/20100101 Firefox/85.0",
    }

    def __init__(
        self,
        auto_compress: bool = False,
        add_watermark: bool = False,
        ignore_cache: bool = False,
        conf: Optional[Config] = None,
        timeout: Optional[float] = None,
        quiet: bool = False,
    ):
        super().__init__(
            auto_compress, add_watermark, ignore_cache, conf, timeout, quiet
        )

        self.cookie: Optional[str] = None
        self.token: Optional[str] = None
        if self.auth_info:
            self.cookie = self.auth_info["cookie"]
            self.token = self.auth_info["token"]
            self.username = self.auth_info["username"]
            self.__headers["Cookie"] = self.cookie

    def login(self, username: str, password: str) -> bool:
        url = self._url("login")
        auth_token, cookie = self._parse_auth_token()

        if not auth_token or not cookie:
            raise MissingAuth("auth token or cookie is None")

        headers = self.__headers.copy()
        headers["Cookie"] = cookie

        data = {
            "login-subject": username,
            "password": password,
            "auth_token": auth_token,
        }

        resp = requests.post(
            url, headers=headers, data=data, allow_redirects=False, timeout=self.timeout
        )
        if resp.status_code == 301:
            # If there is a KEEPLOGIN field in the cookie,
            # pictures will be uploaded as a normal user
            self.cookie = cookie + "; " + resp.headers["Set-Cookie"].split("; ")[0]

            # If there is no KEEPLOGIN field in the cookie,
            # pictures will be uploaded as a tourist
            # self.cookie = cookie

            auth_info = {
                "token": auth_token,
                "cookie": self.cookie,
                "username": username,
                "password": password,
            }
            self._save_auth_info(auth_info)
            return True

        return False

    def _parse_auth_token(self) -> Tuple[Optional[str], Optional[str]]:
        url = self._url("login")
        resp = requests.get(url, headers=self.__headers, timeout=self.timeout)
        if resp.status_code == 200:
            auth_token = re.search(
                r'PF.obj.config.auth_token = "([a-f0-9]{40})"', resp.text
            )
            if not auth_token:
                return None, None
            resp_set_cookie = resp.headers["Set-Cookie"].split("; ")[0]
            return auth_token.group(1), resp_set_cookie
        else:
            logger.error("response error", status_code=resp.status_code)
            return None, None

    @property
    def headers(self):
        assert self.cookie != None

        headers = self.__headers.copy()
        headers.update(
            {
                "Cookie": self.cookie,
            }
        )
        return headers

    def _update_auth_token(self):
        resp = requests.get(self.base_url, headers=self.headers, timeout=self.timeout)
        auth_token = re.search(
            r'PF.obj.config.auth_token = "([a-f0-9]{40})"', resp.text
        )

        assert auth_token != None

        auth_token = auth_token.group(1)
        if not self.auth_info:
            self.auth_info = {}

        self.auth_info["token"] = self.token = auth_token
        self._save_auth_info(self.auth_info)

    def __upload(
        self, image: ImageType, retries: int = 0
    ) -> Union[str, UploadErrorResponse]:
        url, md5, ok = self._check_cache(image)  # type: ignore

        if ok and not self.ignore_cache:
            return url

        self.check_login()

        image = self._compress_image(image)

        if isinstance(image, Path):
            image = self._add_watermark(image)

        logger.debug("uploading", image_path=image)

        url = self._url("json")
        filename_with_suffix = os.path.basename(str(image))
        filename_without_suffix, suffix = os.path.splitext(filename_with_suffix)
        if isinstance(image, Path):
            if suffix.lower() == ".apng":
                suffix = ".png"
            filename = filename_without_suffix + suffix
        else:
            filename = filename_with_suffix + "." + image.mime_type

        file = File("source", image, filename=filename)

        timestamp = int(time.time() * 1000)

        data = {
            "type": "file",
            "action": "upload",
            "timestamp": str(timestamp),
            "auth_token": self.token,
            "nsfw": "0",
        }

        try:
            if not self.quiet:
                resp = upload_with_progress_bar(
                    url, file, self.timeout, data, self.headers
                )
            else:
                resp = requests.post(
                    url,
                    headers=self.headers,
                    data=data,
                    files=file.to_dict(),
                    timeout=self.timeout,
                )
        except requests.exceptions.ConnectionError as e:
            return UploadErrorResponse(400, str(e), str(image))

        resp.encoding = "utf-8"

        try:
            json_resp = resp.json()
        except json.decoder.JSONDecodeError:
            return UploadErrorResponse(resp.status_code, resp.text, str(image))

        try:
            uploaded_url: str = json_resp["image"]["image"]["url"]
            logger.info(
                "uploaded url",
                target=image,
                url=uploaded_url,
            )

            self.cache.save(
                md5,
                IMAGE_BEDS_NAME[self.image_bed_code],
                uploaded_url,
                self.ignore_cache,
            )

            return uploaded_url
        except KeyError:
            if json_resp["error"]["message"] == "请求被拒绝 (auth_token)":
                if retries >= 3:
                    return UploadErrorResponse(
                        resp.status_code, resp.json()["error"]["message"], str(image)
                    )

                logger.warning(
                    "`auth_token` has expired, the program will try to update `auth_token` automatically, number of retries",
                    retries=retries + 1,
                )
                self._update_auth_token()

                return self.__upload(image, retries + 1)
            else:
                return UploadErrorResponse(
                    resp.status_code, resp.json()["error"]["message"], str(image)
                )

    def upload_image(self, image_path: ImagePath) -> Union[str, UploadErrorResponse]:
        return self.__upload(image_path)

    def upload_image_stream(
        self, image: ImageStream
    ) -> Union[str, UploadErrorResponse]:
        return self.__upload(image)

    def get_all_images(self) -> Union[List[ImgtuResponse], ErrorResponse]:
        self.check_login()

        url = self._url(self.username)

        # 集合去重
        images: Set[Tuple[str, ...]] = set()

        def visit_next_page(url: str):
            resp = requests.get(url, headers=self.__headers, timeout=self.timeout)
            resp.encoding = "utf-8"

            if resp.status_code != 200:
                return ErrorResponse(resp.status_code, resp.text)

            # 只获取默认的公开相册里的图片
            # images_object = re.findall(
            #     r"data-privacy=\"public\" data-object='(.+?)'", resp.text)
            # 获取用户所有图片
            images_object = re.findall(r"data-object='(.+?)'", resp.text)
            images_object = [json.loads(parse.unquote(i)) for i in images_object]
            for image in images_object:
                images.add(
                    (
                        image["url"],
                        image["display_url"],
                        image["id_encoded"],
                        image["width"],
                        image["height"],
                    )
                )

            next_page_url = re.search(
                r'data-pagination="next" href="(.+?)" ><span', resp.text
            )
            if next_page_url:
                next_page_url = next_page_url.group(1)
            else:
                next_page_url = ""

            if next_page_url:
                visit_next_page(next_page_url)

        resp = visit_next_page(url)
        if resp:
            return resp

        result: List[ImgtuResponse] = []

        for item in images:
            result.append(
                ImgtuResponse(
                    item[2],
                    item[0],
                    item[1],
                    int(item[3]),
                    int(item[4]),
                )
            )

        return result

    def delete_image(self, img_id: str, retries: int = 0) -> Optional[ErrorResponse]:
        self.check_login()

        url = self._url("json")
        data = {
            "auth_token": self.token,
            "action": "delete",
            "single": "true",
            "delete": "image",
            "deleting[id]": img_id,
        }
        resp = requests.post(
            url, headers=self.__headers, data=data, timeout=self.timeout
        )
        if resp.status_code == 400:
            if resp.json()["error"]["message"] == "请求被拒绝 (auth_token)":
                if retries >= 3:
                    return ErrorResponse(
                        resp.status_code, resp.json()["error"]["message"]
                    )

                logger.warn(
                    "`auth_token` has expired, the program will try to update `auth_token` automatically, number of retries: %d",
                    retries + 1,
                )
                self._update_auth_token()
                return self.delete_image(img_id, retries + 1)

        if resp.status_code == 200:
            return None

        return ErrorResponse(resp.status_code, resp.json()["error"]["message"])

    def delete_images(
        self, imgs_id: List[str], retries: int = 0
    ) -> Dict[str, ErrorResponse]:
        self.check_login()

        url = self._url("json")
        data = {
            "auth_token": self.token,
            "action": "delete",
            "from": "list",
            "multiple": "true",
            "delete": "images",
            "deleting[ids][]": imgs_id,
        }
        resp = requests.post(
            url, headers=self.__headers, data=data, timeout=self.timeout
        )
        json_resp = resp.json()
        if resp.status_code == 400:
            if json_resp["error"]["message"] == "请求被拒绝 (auth_token)":
                if retries >= 3:
                    # 删除多张图片时如果重试3次仍无法成功响应则退出程序
                    logger.fatal(
                        "authentication information is invalid",
                        error=resp.json()["error"]["message"],
                    )

                logger.warn(
                    "`auth_token` has expired, the program will try to update `auth_token` automatically, number of retries: %d",
                    retries + 1,
                )
                self._update_auth_token()
                return self.delete_images(imgs_id, retries + 1)
            elif json_resp["error"]["code"] == 106:
                # imgtu 只有删除的所有 id 都无效时才会返回这个错误
                # 只要有一个有效 id，就会返回 200
                return {img_id: ErrorResponse(404, "服务器中无此图片") for img_id in imgs_id}

        return {}

    def _url(self, path: str) -> str:
        return self.base_url + path

    def __repr__(self):
        return "imgtu.com"

    @property
    def description(self):
        return "中国大陆访问可能不太稳定"



#!/usr/bin/env python3
# -*- coding:utf-8 -*-

import re
import requests

from typing import List, Optional, Dict, Any, Union
from pathlib import Path
from up2b.up2b_lib.custom_types import (
    Config,
    ErrorResponse,
    ImageBedType,
    ImagePath,
    ImageStream,
    ImageType,
    SMMSResponse,
    UploadErrorResponse,
)
from up2b.up2b_lib.file import File
from up2b.up2b_lib.up2b_api import Base
from up2b.up2b_lib.constants import IMAGE_BEDS_NAME, ImageBedCode
from up2b.up2b_lib.log import child_logger
from up2b.up2b_lib.http import upload_with_progress_bar

logger = child_logger(__name__)


class SM(Base):
    image_bed_type = ImageBedType.common
    image_bed_code = ImageBedCode.SM_MS
    base_url = "https://smms.app/api/v2/"
    max_size = 5 * 1024 * 1024

    def __init__(
        self,
        auto_compress: bool = False,
        add_watermark: bool = False,
        ignore_cache: bool = False,
        conf: Optional[Config] = None,
        timeout: Optional[float] = None,
        quiet: bool = False,
    ):
        super().__init__(
            auto_compress, add_watermark, ignore_cache, conf, timeout, quiet
        )

        if self.auth_info:
            self.token: str = self.auth_info["token"]

            self.headers = {
                "Authorization": self.token,
            }

    def login(self, username: str, password: str) -> bool:
        try:
            token = self._get_api_token(username, password)
            self._save_auth_info(
                {"token": token, "username": username, "password": password}
            )
            return True
        except Exception as e:
            logger.error(e.__str__())
            return False

    def _auto_login(self):
        if not self.auth_info:
            raise ValueError("auth info is not found")

        username = self.auth_info["username"]
        password = self.auth_info["password"]
        self.login(username, password)
        self.auth_info = self._read_auth_info()

        assert self.auth_info is not None

    def _get_api_token(self, username: str, password: str) -> str:
        url = self._url("token")
        data = {"username": username, "password": password}
        resp = requests.post(url, data=data, timeout=self.timeout).json()
        return resp["data"]["token"]

    def _get_user_profile(self) -> Optional[Dict[str, str]]:
        url = self._url("profile")
        resp = requests.post(url, headers=self.headers, timeout=self.timeout).json()
        if resp["success"]:
            return resp["data"]
        else:
            if self._login_expired(resp):
                self._auto_login()
                self.token = self.auth_info["token"]  # type: ignore
        return None

    def __upload(self, image: ImageType, retries=0) -> Union[str, UploadErrorResponse]:
        self.check_login()

        url, md5, ok = self._check_cache(image)  # type: ignore

        if ok and not self.ignore_cache:
            return url

        logger.debug("uploading", image=image)

        image = self._compress_image(image)

        if isinstance(image, Path):
            image = self._add_watermark(image)

        # sm.ms不管出不出错，返回的状态码都是200
        url = self._url("upload")

        file = File("smfile", image)

        if self.quiet:
            res = requests.post(
                url, headers=self.headers, files=file.to_dict(), timeout=self.timeout  # type: ignore
            )
        else:
            res = upload_with_progress_bar(url, file, self.timeout, None, self.headers)

        logger.debug("响应", status=res.status_code, body=res.text)

        resp = res.json()

        if resp["success"] or resp["code"] == "image_repeated":
            uploaded_url: str = resp["data"]["url"]
            logger.info(
                "uploaded",
                image=image,
                url=uploaded_url,
            )

            self.cache.save(
                md5,
                IMAGE_BEDS_NAME[self.image_bed_code],
                uploaded_url,
                self.ignore_cache,
            )
            return uploaded_url
        else:
            if resp["code"] == "image_repeated":
                # 如果图片重复，会返回重复的图片的链接，所以此处不报错
                logger.info("repeated image", image=resp["images"])
                image_url: str = resp["images"]
                return image_url
            elif self._login_expired(resp):
                assert self.auth_info is not None

                if retries >= 3:
                    return UploadErrorResponse(401, "认证信息无效", str(image))

                logger.warn(
                    "`auth_token` has expired, the program will try to update `auth_token` automatically, number of retries: %d",
                    retries + 1,
                )

                self._auto_login()
                assert self.auth_info is not None
                self.token = self.auth_info["token"]

                return self.__upload(image, retries + 1)
            else:
                error = resp["message"]
                logger.error(
                    "upload failed",
                    image=image,
                    error=error,
                )
                return UploadErrorResponse(400, error, str(image))

    def upload_image(self, image_path: ImagePath):
        return self.__upload(image_path)

    def upload_image_stream(self, image: ImageStream):
        return self.__upload(image)

    def history(self) -> Dict[str, Any]:
        """
        Temporary History - IP Based Temporary Upload History
        """
        self.check_login()

        url = self._url("history")
        resp = requests.get(url, headers=self.headers, timeout=self.timeout)
        return resp.json()

    def clear(self) -> Dict[str, Any]:
        """
        Clear Temporary History - Clear IP Based Temporary Upload History
        """
        self.check_login()

        url = self._url("clear")
        resp = requests.get(url, headers=self.headers, timeout=self.timeout)
        return resp.json()

    def upload_history(self) -> Union[ErrorResponse, List[Dict[str, Any]]]:
        self.check_login()

        url = self._url("upload_history")
        resp = requests.get(url, headers=self.headers, timeout=self.timeout).json()
        if resp["code"] != "success":
            if resp["code"] == "unauthorized":
                return ErrorResponse(401, resp["message"])
            else:
                return ErrorResponse(0, resp)
        return resp["data"]

    def get_all_images(self) -> Union[ErrorResponse, List[SMMSResponse]]:
        self.check_login()

        images: List[SMMSResponse] = []
        data = self.upload_history()
        if isinstance(data, ErrorResponse):
            return data

        for file in data:
            images.append(
                SMMSResponse(
                    file["url"],
                    file["delete"],
                    file["width"],
                    file["height"],
                )
            )

        return images

    def delete_image(self, delete_url: str) -> Optional[ErrorResponse]:
        self.check_login()

        resp = requests.get(delete_url, timeout=self.timeout)

        re_res = re.search(r'<div class="card-body">\n\s+(.*?)\n', resp.text)
        if not re_res:
            return ErrorResponse(500, "未知错误")

        msg = re_res.group(1)
        if msg == "File is deleted and our cache will refresh within minutes.":
            return ErrorResponse(400, "图片已被删除，请耐心等待服务器刷新缓存")

        if msg == "Picture was already deleted.":
            return ErrorResponse(404, "图片已被删除，请勿重复删除")
        else:
            return ErrorResponse(0, msg)

    def delete_images(self, urls: List[str]) -> Dict[str, ErrorResponse]:
        self.check_login()

        result: Dict[str, ErrorResponse] = {}
        for url in urls:
            r = self.delete_image(url)
            if r:
                result[url] = r
        return result

    def get_all_images_in_image_bed(self) -> List[str]:  # type: ignore
        pass

    def _url(self, path: str) -> str:
        return self.base_url + path

    def _login_expired(self, resp: dict):
        return resp["message"] == "Get user profile failed, invalid Authorization."

    def __repr__(self):
        return "sm.ms"

    @property
    def description(self):
        return "中国大陆访问可能不太稳定"



from pathlib import Path
import requests

from abc import ABC, abstractmethod
from typing import overload, Optional, List, Tuple, Dict, Union
from up2b.up2b_lib.cache import Cache
from up2b.up2b_lib.constants import ImageBedCode
from up2b.up2b_lib.custom_types import (
    Config,
    DownloadErrorResponse,
    ErrorResponse,
    GitGetAllImagesResponse,
    ImageBedType,
    ImagePath,
    ImageStream,
    ImageType,
    AuthInfo,
    ImgtuResponse,
    SMMSResponse,
    UploadErrorResponse,
)

def choose_image_bed(image_bed_code: int) -> None: ...

AllImagesResponse = (
    List[GitGetAllImagesResponse] | List[SMMSResponse] | List[ImgtuResponse]
)

class ImageBedAbstract(ABC):
    @abstractmethod
    def get_all_images(self) -> Union[AllImagesResponse, ErrorResponse]: ...
    @abstractmethod
    def upload_image(self, image_path: ImagePath) -> Union[str, UploadErrorResponse]:
        pass
    @abstractmethod
    def upload_image_stream(
        self, image: ImageStream
    ) -> Union[str, UploadErrorResponse]: ...
    @overload
    @abstractmethod
    def delete_image(
        self, unique_id: str, retries: int = ...
    ) -> Optional[ErrorResponse]: ...
    @overload
    @abstractmethod
    def delete_image(
        self,
        sha: str,
        url: str,
        message: str = ...,
    ) -> Optional[ErrorResponse]: ...
    @overload
    @abstractmethod
    def delete_images(
        self, unique_ids: List[str], retries: int = ...
    ) -> Dict[str, ErrorResponse]: ...
    @overload
    @abstractmethod
    def delete_images(
        self,
        info: List[Tuple[str, str]],
        message: str = ...,
    ) -> Dict[str, ErrorResponse]: ...
    @property
    @abstractmethod
    def description(self) -> str: ...

Images = Tuple[Union[ImageType, DownloadErrorResponse], ...]

class Base:
    image_bed_code: ImageBedCode
    max_size: int
    conf: Config
    image_bed_type: ImageBedType
    conf_file: str
    auth_info: Optional[AuthInfo]
    add_watermark: bool
    auto_compress: bool
    ignore_cache: bool
    quiet: bool
    timeout: float
    cache: Cache

    def __init__(
        self,
        auto_compress: bool = ...,
        add_watermark: bool = ...,
        ignore_cache: bool = ...,
        conf: Optional[Config] = ...,
        timeout: Optional[float] = ...,
        quiet: bool = ...,
    ) -> None: ...
    def check_login(self) -> None: ...
    def _read_auth_info(self) -> Optional[AuthInfo]: ...
    def _save_auth_info(self, auth_info: Dict[str, str]) -> None: ...
    def _exceed_max_size(self, images: Images) -> Tuple[bool, Optional[str]]: ...
    def _check_images_valid(self, images: Images): ...
    def _compress_image(self, image: ImageType) -> ImageType: ...
    def _add_watermark(self, image_path: ImagePath) -> ImagePath: ...
    def _clear_cache(self) -> None: ...
    def _check_cache(self, image: Path) -> Tuple[str, str, bool]: ...
    def upload_images(
        self, *images: Union[ImageType, DownloadErrorResponse], to_console: bool = ...
    ) -> List[Union[str, UploadErrorResponse]]: ...

class GitBase(Base, ImageBedAbstract):
    headers: Dict[str, str]
    api_url: str

    token: str
    username: str
    repo: str
    folder: str

    def __init__(
        self,
        auto_compress: bool = ...,
        add_watermark: bool = ...,
        ignore_cache: bool = ...,
        conf: Optional[Config] = ...,
        timeout: Optional[float] = None,
        quiet: bool = ...,
    ) -> None: ...
    def login(
        self, token: str, username: str, repo: str, folder: str = ...
    ) -> None: ...
    def _upload(
        self, image: ImageType, data: Dict[str, str], request_method: str = ...
    ) -> Union[str, UploadErrorResponse]: ...
    @abstractmethod
    def _get_all_images_in_image_bed(
        self,
    ) -> requests.Response: ...
    def get_all_images(self) -> Union[List[GitGetAllImagesResponse], ErrorResponse]: ...
    def _delete_image(
        self,
        sha: str,
        url: str,
        message: str = ...,
        extra: Optional[Dict[str, str]] = ...,
    ) -> Optional[ErrorResponse]: ...
    def delete_images(
        self,
        info: List[Tuple[str, str]],
        message: str = ...,
    ) -> Dict[str, ErrorResponse]: ...
    @property
    def base_url(self) -> str: ...



#!/usr/bin/env python3
# -*- coding:utf-8 -*-

import os
import time
import json
import shutil
import requests

from io import BytesIO
from abc import ABC, abstractmethod
from typing import Callable, Optional, List, Tuple, Dict, Union
from pathlib import Path
from up2b.up2b_lib.cache import Cache
from up2b.up2b_lib.constants import (
    CONF_FILE,
    CACHE_PATH,
    IMAGE_BEDS_NAME,
    PYTHON_VERSION,
    ImageBedCode,
)

if PYTHON_VERSION >= (3, 8):
    from functools import cached_property
else:
    cached_property = property

from up2b.up2b_lib.custom_types import (
    Config,
    DownloadErrorResponse,
    ErrorResponse,
    GitGetAllImagesResponse,
    ImageBedType,
    ImagePath,
    ImageStream,
    ImageType,
    AuthInfo,
    UploadErrorResponse,
    WaterMarkConfig,
)
from up2b.up2b_lib.log import child_logger
from up2b.up2b_lib.utils import check_image_exists, read_conf, timeout_in_env
from up2b.up2b_lib.errors import UnsupportedType, OverSizeError

logger = child_logger(__name__)


def choose_image_bed(image_bed_code: int):
    if type(image_bed_code) != int:
        raise TypeError(
            "image bed code must be an integer, not %s" % str(type(image_bed_code))
        )
    try:
        with open(CONF_FILE, "r+") as f:
            conf = json.loads(f.read())
            f.seek(0, 0)
            conf["image_bed"] = image_bed_code
            f.write(json.dumps(conf))
            f.truncate()
    except FileNotFoundError:
        with open(CONF_FILE, "w") as f:
            f.write(json.dumps({"image_bed": image_bed_code}))


class ImageBedAbstract(ABC):
    @abstractmethod
    def get_all_images(self):
        pass

    @abstractmethod
    def upload_image(self, image_path: ImagePath) -> Union[str, UploadErrorResponse]:
        pass

    @abstractmethod
    def upload_image_stream(
        self, image: ImageStream
    ) -> Union[str, UploadErrorResponse]:
        pass

    @abstractmethod
    def delete_image(self, *args, **kwargs) -> Optional[ErrorResponse]:
        pass

    @abstractmethod
    def delete_images(self, *args, **kwargs) -> Dict[str, ErrorResponse]:
        pass

    @property
    @abstractmethod
    def description(self) -> str:
        pass


class Base(ImageBedAbstract):
    image_bed_code: ImageBedCode
    max_size: int
    conf: Config
    image_bed_type: ImageBedType
    cache = Cache()

    def __init__(
        self,
        auto_compress: bool = False,
        add_watermark: bool = False,
        ignore_cache: bool = False,
        conf: Optional[Config] = None,
        timeout: Optional[float] = None,
        quiet: bool = False,
    ):
        self.timeout = timeout_in_env() if timeout is None else timeout
        self.quiet = quiet
        self.conf = conf if conf != None else read_conf()

        self.auth_info: Optional[AuthInfo] = self._read_auth_info()
        self.add_watermark: bool = add_watermark
        if self.add_watermark:
            if not self.conf.watermark:
                logger.fatal(
                    "you have enabled the function of adding watermark, but the watermark is not configured, please configure the text watermark through `--config-text-watermark`"
                )
        self.auto_compress: bool = auto_compress
        self.ignore_cache: bool = ignore_cache

    def check_login(self):
        if not self.auth_info:
            logger.fatal(
                "you have not logged in yet, please log in first.",
                code=self.image_bed_code,
                name=self,
            )

    def _read_auth_info(self) -> Optional[AuthInfo]:
        auth_data = self.conf.auth_data
        if not auth_data:
            return

        assert isinstance(auth_data, dict)

        auth_info = auth_data.get(self.image_bed_code)

        assert auth_info is None or isinstance(auth_info, dict)

        return auth_info

    def _save_auth_info(self, auth_info: Dict[str, str]):
        logger.debug("current image bed code", code=self.image_bed_code)
        try:
            with open(CONF_FILE, "r+") as f:
                conf = json.loads(f.read())
                try:
                    conf["auth_data"][str(self.image_bed_code.value)] = auth_info
                except KeyError:
                    conf["auth_data"] = {}
                    conf["auth_data"][str(self.image_bed_code.value)] = auth_info
                f.seek(0, 0)
                f.write(json.dumps(conf))
                f.truncate()
        except FileNotFoundError:
            logger.fatal(
                "auth configure file is not found, please choose image bed with `--choose-site` or `-c` first."
            )
        except Exception as e:
            logger.fatal("save auth configure failed", error=e)

    def _exceed_max_size(
        self, images: Tuple[Union[ImageType, DownloadErrorResponse], ...]
    ) -> Tuple[int, Optional[str]]:
        for img in images:
            if isinstance(img, DownloadErrorResponse):
                continue

            size = os.path.getsize(img) if isinstance(img, Path) else len(img.stream)
            if size > self.max_size:
                return size, str(img) if isinstance(img, Path) else img.filename

        return 0, None

    def _check_images_valid(
        self, images: Tuple[Union[ImageType, DownloadErrorResponse], ...]
    ):
        """
        Check if all images exceed the max size or can be compressed
        """
        if not self.auto_compress:
            size, img = self._exceed_max_size(images)
            if size:
                raise OverSizeError(
                    f"{img}: {size /1024/1024:.2f}M > {self.max_size/1024/1024}M"
                )
        else:
            for _img in images:
                if isinstance(_img, DownloadErrorResponse):
                    continue

                mime_type = (
                    _img.suffix.lower() if isinstance(_img, Path) else _img.mime_type
                )
                if mime_type not in ["jpg", "png", "jpeg"]:
                    raise UnsupportedType(
                        "currently does not support compression of this type of image: %s"
                        % mime_type.upper()
                    )

    def _compress_image(self, image: ImageType) -> ImageType:
        if not self.auto_compress:
            return image

        logger.debug("compressing image", image=image)

        try:
            from PIL import Image
        except ModuleNotFoundError:
            logger.fatal(
                "you have enabled the automatic image compression feature, but [ pillow ] is not installed, please execute `pip install pillow` before enabling this feature"
            )

        raw_size = (
            os.path.getsize(image) if isinstance(image, Path) else len(image.stream)
        )
        if raw_size > self.max_size:
            filename = os.path.basename(str(image)).split(".")[0]

            scale = self.max_size / raw_size
            img = Image.open(image if isinstance(image, Path) else image.stream)

            def compress(img: Image.Image, scale: float) -> BytesIO:
                format = img.format
                width, height = img.size
                img_io = BytesIO()

                if format == "PNG":
                    img = img.convert("RGB")
                    img.format = "JPEG"  # type: ignore
                    img.save(img_io, "jpeg")
                elif format == "JPEG":
                    img = img.resize(
                        (int(width * scale), int(height * scale)), Image.ANTIALIAS
                    )
                    img.format = "JPEG"  # type: ignore
                    img.save(img_io, "jpeg")
                else:
                    raise UnsupportedType(
                        "currently does not support compression of this type of image: %s"
                        % format
                    )

                if img_io.tell() > self.max_size:
                    scale = self.max_size / img_io.tell()
                    return compress(img, scale)
                return img_io

            if img.format == "PNG":
                filename += ".jpeg"
            else:
                filename += "." + img.format.lower()  # type: ignore
            img_io = compress(img, scale)

            if not os.path.exists(CACHE_PATH):
                os.mkdir(CACHE_PATH)

            compressed_size = img_io.tell()

            img_cache_path = CACHE_PATH / filename
            with img_cache_path.open("wb") as f:
                f.write(img_io.getbuffer())

            logger.info(
                "image compression complete",
                image=img_cache_path,
                raw_size=f"{raw_size / 1024}b",
                compressed_size=f"{compressed_size / 1024}k",
            )

            return img_cache_path

        return image

    def _add_watermark(self, image_path: ImagePath) -> ImagePath:
        if not self.add_watermark:
            return image_path

        from up2b.up2b_lib.watermark import AddWatermark, TypeFont

        conf: WaterMarkConfig = self.conf["watermark"]  # type: ignore

        assert isinstance(conf, dict)

        aw = AddWatermark(conf["x"], conf["y"], conf["opacity"])
        return aw.add_text_watermark(
            image_path,
            [TypeFont(conf["text"], conf["size"], conf["font"], (0, 0, 0))],
        )

    def _clear_cache(self):
        if os.path.exists(CACHE_PATH):
            shutil.rmtree(CACHE_PATH)
            os.mkdir(CACHE_PATH)

            logger.debug("cache folder has been cleared", cache_path=CACHE_PATH)

    def _check_cache(self, image: Path) -> Tuple[str, str, bool]:
        url, md5, ok = self.cache.check_cache_of_image_bed(
            image, IMAGE_BEDS_NAME[self.image_bed_code]
        )

        if ok:
            if self.ignore_cache:
                logger.info("缓存中找到此图片链接，但用户选择忽略缓存强制上传")
            else:
                logger.info("缓存中找到此图片链接", url=url)

            return (url, md5, ok)

        logger.info("缓存中未找到此图片链接，开始上传")
        return (url, md5, ok)

    def upload_images(
        self, *images: Union[ImageType, DownloadErrorResponse], to_console: bool = True
    ) -> List[Union[str, DownloadErrorResponse, UploadErrorResponse]]:
        self.check_login()

        check_image_exists(images)

        self._check_images_valid(images)

        image_urls: List[Union[str, DownloadErrorResponse, UploadErrorResponse]] = []
        for img in images:
            if isinstance(img, DownloadErrorResponse):
                result = img
            elif isinstance(img, Path):
                result = self.upload_image(img)
            else:
                result = self.upload_image_stream(img)

            image_urls.append(result)

        if to_console:
            for iu in image_urls:
                if isinstance(iu, UploadErrorResponse):
                    logger.error("上传出错", status_code=iu.status_code, error=iu.error)
                else:
                    print(iu)

        self._clear_cache()
        return image_urls


class GitBase(Base):
    headers: Dict[str, str]
    api_url: str

    token: str
    username: str
    repo: str
    folder: str

    def __init__(
        self,
        auto_compress: bool = False,
        add_watermark: bool = False,
        ignore_cache: bool = False,
        conf: Optional[Config] = None,
        timeout: Optional[float] = None,
        quiet: bool = False,
    ):
        super().__init__(
            auto_compress, add_watermark, ignore_cache, conf, timeout, quiet
        )

        if self.auth_info:
            self.token = self.auth_info["token"]
            self.username = self.auth_info["username"]
            self.repo = self.auth_info["repo"]
            self.folder = self.auth_info["folder"]

    def login(self, token: str, username: str, repo: str, folder: str = "up2b"):
        auth_info = {
            "token": token,
            "username": username,
            "repo": repo,
            "folder": folder,
        }
        self._save_auth_info(auth_info)

    def _upload(
        self, image: ImageType, data: Dict[str, str], request_method: str = "put"
    ) -> Union[str, UploadErrorResponse]:
        self.check_login()

        url, md5, ok = self._check_cache(image)  # type: ignore

        if ok and not self.ignore_cache:
            return url

        image = self._compress_image(image)
        if isinstance(image, Path):
            image = self._add_watermark(image)

        suffix = os.path.splitext(str(image))[-1]
        if suffix.lower() == ".apng":
            suffix = ".png"
        filename = f"{int(time.time() * 1000)}{suffix}"

        url = self.base_url + filename

        logger.debug("request headers", headers=self.headers)

        resp = requests.request(
            request_method, url, headers=self.headers, json=data, timeout=self.timeout
        )
        if resp.status_code == 201:
            uploaded_url: str = resp.json()["content"]["download_url"]
            logger.info("uploaded", image=image, url=uploaded_url)
            if hasattr(self, "cdn_url") and callable(getattr(self, "cdn_url")):
                return self.cdn_url(uploaded_url)  # type: ignore

            self.cache.save(
                md5,
                IMAGE_BEDS_NAME[self.image_bed_code],
                uploaded_url,
                self.ignore_cache,
            )

            return uploaded_url
        else:
            error = resp.json()["message"]
            logger.error("upload failed", image=image, error=error)
            return UploadErrorResponse(resp.status_code, error, str(image))

    def upload_images(
        self, *images: Union[ImageType, DownloadErrorResponse], to_console: bool = True
    ) -> List[Union[str, DownloadErrorResponse, UploadErrorResponse]]:
        self.check_login()

        check_image_exists(images)

        self._check_images_valid(images)

        image_urls: List[Union[str, DownloadErrorResponse, UploadErrorResponse]] = []
        for img in images:
            if isinstance(img, DownloadErrorResponse):
                result = img
            elif isinstance(img, Path):
                result = self.upload_image(img)
            else:
                result = self.upload_image_stream(img)

            image_urls.append(result)

        if to_console:
            for iu in image_urls:
                print(iu)

        self._clear_cache()
        return image_urls

    @abstractmethod
    def _get_all_images_in_image_bed(
        self,
    ) -> requests.Response:
        pass

    def get_all_images(self) -> Union[List[GitGetAllImagesResponse], ErrorResponse]:
        self.check_login()

        resp = self._get_all_images_in_image_bed()
        if resp.status_code != 200:
            return ErrorResponse(resp.status_code, resp.text)

        all_images_resp: List[Dict[str, str]] = resp.json()
        images: List[GitGetAllImagesResponse] = []
        for file in all_images_resp:
            download_url = file["download_url"]
            if hasattr(self, "cdn_url") and callable(getattr(self, "cdn_url")):
                self.cdn_url: Callable[[str], str]
                download_url: str = self.cdn_url(file["download_url"])
            images.append(
                GitGetAllImagesResponse(
                    download_url,
                    file["sha"],
                    file["url"],
                )
            )
        return images

    def _delete_image(
        self,
        sha: str,
        url: str,
        message: str = "Delete pictures that are no longer used",
        extra: Optional[Dict[str, str]] = None,
    ):
        self.check_login()

        data = {"sha": sha, "message": message}
        if extra:
            data.update(extra)

        resp = requests.delete(
            url, headers=self.headers, json=data, timeout=self.timeout
        )
        if resp.status_code == 200:
            return None

        return ErrorResponse(resp.status_code, resp.json()["message"])

    def delete_images(
        self,
        info: List[Tuple[str, str]],
        message: str = "Delete pictures that are no longer used",
    ) -> Dict[str, ErrorResponse]:
        self.check_login()

        failed: Dict[str, ErrorResponse] = {}
        for sha, url in info:
            result = self.delete_image(sha, url, message)
            if result:
                failed["sha"] = result
        return failed

    @cached_property
    def base_url(self) -> str:
        return "%s/repos/%s/%s/contents/%s/" % (
            self.api_url,
            self.username,
            self.repo,
            self.folder,
        )
